package icfp2001;

import java.util.*;

/**
 * A generic node in the SML tree
 *
 */
public abstract class Node implements Cloneable {

    /**
     * A reference to the parent node, or null if this is a
     * root node
     */
    protected Node mParent;

    /**
     * A list of the children of this node. This is an ordered
     * list
     */
    protected List mChildren;

    /**
     * provided for visitors to cache info
     */
    public HashMap mHash;

    /**
     * Constructs a new node with no parent
     *
     */
    public Node() {
        this(null);
    }

    /**
     * Constructs a new node with the specified parent
     *
     * @param pParent the parent node, or null for a root node
     */
    public Node(Node pParent) {
        mParent = pParent;
        mChildren = new ArrayList();
        mHash = new HashMap();
    }

    /**
     * Returns the parent of this node
     *
     * @return the parent
     */
    public Node getParent() {
        return mParent;
    }

    /**
     * Sets the parent for this node
     *
     * @param pParent the parent to set
     */
    protected void setParent(Node pParent) {
        mParent = pParent;
    }

    /**
     * Gets the number of children of this node
     *
     * @return the number of children under this node
     */
    public int numChildren() {
        return mChildren.size();
    }

    /**
     * Returns an iterator over the children of this node
     *
     * @return an iterator over the children of this node
     */
    public Iterator childIterator() {
        return mChildren.iterator();
    }

    /**
     * Returns the list backing the children of this node. Changing 
     * this List will change the children of this Node.
     *
     * @return the List object containing the Node objects
     * that are children of this Node. 
     */
    protected List getChildren() {
        return mChildren;
    }

    public Node getChild(int pIndex) 
    {

        return (Node)mChildren.get(pIndex);
    }

    /**
     * Adds a child to this node
     *
     * @param pChild the child to add
     */
    public void addChild(Node pChild) {
        mChildren.add(pChild);
        pChild.setParent(this);
    }

    /**
     * Sets a specific child below this node, replacing
     * the one that was there
     * 
     * @param pChildIndex the index of the child to replace.
     * @param pNewChild the new child
     *
     * @return the old child
     */
    public Node setChild(int pChildIndex, Node pNewChild) {
        Node vOldChild = (Node)mChildren.get(pChildIndex);
        mChildren.set(pChildIndex, pNewChild);
        vOldChild.setParent(null);
        return vOldChild;
    }

    /**
     * Inserts a child at the specified index
     *
     * @param pChildIndex the index of the child to be entered
     * @param pNewChild the new child
     */
    public void insertChild(int pChildIndex, Node pNewChild) {
        mChildren.add(pChildIndex, pNewChild);
        pNewChild.setParent(this);
    }

    /**
     * Adds a List of Children to this Node
     *
     * @param pChildList a list of children to add
     */
    public void addChildren(List pChildList) {
        Iterator vChildIt = pChildList.iterator();
        Node vChild;
        while (vChildIt.hasNext()) {
            vChild = (Node)vChildIt.next();

            addChild(vChild);
        }
    }

    /**
     * Finds the index of a child
     *
     *@param pChild the child to find
     *@return the index, or -1 if not found
     */
    public int getChildIndex(Node pChild)
    {
        for (int i = 0; i < numChildren(); i++)
        {
            if (mChildren.get(i).equals(pChild))
            {
                return i;
            }
        }

        return -1;
    }
     

    /**
     * Accepts a NodeVisitor
     *
     * @pVisitor the visitor
     *
     * @return the return value from the visit
     */
    public abstract Object accept(NodeVisitor pVisitor);


    /**
     * Deletes a node from this node's children list
     *
     * @param pNode the child to delete
     */
    public void deleteChild(Node pNode)
    {
        mChildren.remove(pNode);
        if (pNode.getParent().equals(this))
        {
            pNode.setParent(null);
        }
    }

    /**
     * Detach this node from its parent
     *
     */
    public void detach()
    {
        Node vParent = getParent();
        if (vParent != null)
        {
            vParent.deleteChild(this);
        }
    }

    public Object clone()
    {
        try
        {
            return super.clone();
        }
        catch (Exception e)
        {
            throw new InternalError("Couldn't clone");
        }
    }

    public Node detachedCopy()
    {
        try
        {
            Node vClone = (Node)super.clone();
            vClone.mHash = new HashMap();
            vClone.setParent(null);
            vClone.mChildren = new ArrayList();

            return vClone;
        }
        catch (CloneNotSupportedException vEx)
        {
            System.err.println("ERROR in clone!");
            return null;
        }
    }
    
    /**
     * returns a deep copy of this node
     */
    public Node deepCopy()
    {
        try
        {
            Node vClone = (Node)super.clone();
            vClone.mHash = new HashMap();
            vClone.setParent(null);

            Iterator vIt = vClone.childIterator();
            vClone.mChildren = new ArrayList();
            Node vChild;
            while(vIt.hasNext())
            {
                vChild=(Node)vIt.next();

                vClone.addChild(vChild.deepCopy());
            }

            return vClone;
        }
        catch (CloneNotSupportedException vEx)
        {
            System.err.println("ERROR in clone!");
            return null;
        }
    }

    public int cost()
    {
        return 0;
    }


    public Context getEntryContext()
    {
        return (Context)mHash.get(Context.ENTRY_CONTEXT);
    }

    public void setEntryContext(Context pContext)
    {
        Context vContext = null;
        if (pContext != null)
        {
            vContext = (Context)pContext.clone();
        }
        mHash.put(Context.ENTRY_CONTEXT, vContext);
    }
 }
 
